#:coding:utf-8

import json
import requests
from subprocess import call as sbcall
from subprocess import check_output as sbcheck_output
import pandas as pd
import numpy as np
import os
from tqdm import tqdm
from functools import reduce
from time import sleep

from NORDic.NORDic_DS.get_drug_signatures import drugname2pubchem, pubchem2drugname
from NORDic.UTILS.LINCS_utils import get_user_key, build_url, post_request

def retrieve_drug_targets(file_folder, drug_names, TARGET_args={}, gene_list=[], sources=["DrugBank","MINERVA","LINCS","TTD","DrugCentral"], quiet=False):
    '''
    Retrieve drug targets from several online sources

    ...

    Parameters
    ----------
    drug_names : Python character string list
        list of common drug names
    TARGET_args : Python dictionary
        [default={}] : see for each source
    gene_list : Python character string list
        [default=[]] : list of HGNC symbols
    sources : Python character string list
        [default=["DrugBank","MINERVA","LINCS","TTD","DrugCentral"]] : list of source names (in ["DrugBank","MINERVA","LINCS","TTD","DrugCentral"])
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    targets_df : Pandas DataFrame
        rows/[genes] x columns/[drug names], values are the number of times a target is connected to a drug across all queried databases
    '''
    assert all([s in ["DrugBank","MINERVA","LINCS","TTD","DrugCentral"] for s in sources])
    targets_list = []

    if ("MINERVA" in sources):
        if (not quiet):
            print("<DRUG_TARGETS> MINERVA")
        if (not os.path.exists(file_folder+"drug_targets_MINERVA.csv")):
           targets_MINERVA = get_targets_MINERVA(drug_names, quiet=quiet)
           targets_MINERVA.to_csv(file_folder+"drug_targets_MINERVA.csv")
           print(targets_MINERVA.head())
        targets_MINERVA = pd.read_csv(file_folder+"drug_targets_MINERVA.csv", index_col=0)
        if (targets_MINERVA.shape[1]<len(drug_names)):
            targets_MINERVA = pd.concat(tuple([targets_MINERVA]+[pd.DataFrame([],index=targets_MINERVA.index,columns=[d]) for d in drug_names if (d not in targets_MINERVA.columns)]), axis=1).fillna("")
        targets_list.append(targets_MINERVA[drug_names])

    if ("DrugBank" in sources):
        if (not quiet):
            print("<DRUG_TARGETS> DrugBank")
        if ("DrugBank" in TARGET_args and all([x in TARGET_args["DrugBank"] for x in ["path_to_drugbank","target_fname","drug_fname"]])):
            if (not os.path.exists(file_folder+"drug_targets_DrugBank.csv")):
                targets_DrugBank = get_targets_DrugBank(drug_names, **TARGET_args["DrugBank"], quiet=quiet)
                targets_DrugBank.to_csv(file_folder+"drug_targets_DrugBank.csv")
                print(targets_DrugBank.head())
            targets_DrugBank = pd.read_csv(file_folder+"drug_targets_DrugBank.csv", index_col=0)
            if (targets_DrugBank.shape[1]<len(drug_names)):
                targets_DrugBank = pd.concat(tuple([targets_DrugBank]+[pd.DataFrame([],index=targets_DrugBank.index,columns=[d]) for d in drug_names if (d not in targets_DrugBank.columns)]), axis=1).fillna("")
            targets_list.append(targets_DrugBank[drug_names])
        else:
            print("One of the following fields: %s is not present in input" % str(["path_to_drugbank","target_fname","drug_fname"]))

    if ("LINCS" in sources):
        if (not quiet):
            print("<DRUG_TARGETS> LINCS")
        if ("LINCS" in TARGET_args):
            if (not os.path.exists(file_folder+"drug_targets_LINCS.csv")):
                targets_LINCS = get_targets_LINCS(drug_names, **TARGET_args["LINCS"], quiet=quiet)
                targets_LINCS.to_csv(file_folder+"drug_targets_LINCS.csv")
                print(targets_LINCS.head())
            targets_LINCS = pd.read_csv(file_folder+"drug_targets_LINCS.csv", index_col=0)
            if (targets_LINCS.shape[1]<len(drug_names)):
                targets_LINCS = pd.concat(tuple([targets_LINCS]+[pd.DataFrame([],index=targets_LINCS.index,columns=[d]) for d in drug_names if (d not in targets_LINCS.columns)]), axis=1).fillna("")
            targets_list.append(targets_LINCS[drug_names])
        else:
            print("One of the following fields: %s is not present in input" % str(["path_to_lincs","credentials"]))

    if ("TTD" in sources):
        if (not quiet):
            print("<DRUG_TARGETS> TTD")

        if (not os.path.exists(file_folder+"drug_targets_TTD.csv")):
            targets_TTD = get_targets_TTD(drug_names, quiet=quiet)
            targets_TTD.to_csv(file_folder+"drug_targets_TTD.csv")
            print(targets_TTD.head())
        targets_TTD = pd.read_csv(file_folder+"drug_targets_TTD.csv", index_col=0)
        if (targets_TTD.shape[1]<len(drug_names)):
            targets_TTD = pd.concat(tuple([targets_TTD]+[pd.DataFrame([],index=targets_TTD.index,columns=[d]) for d in drug_names if (d not in targets_TTD.columns)]), axis=1).fillna("")
        targets_list.append(targets_TTD[drug_names])

    if ("DrugCentral" in sources):
        if (not quiet):
            print("<DRUG_TARGETS> DrugCentral")
        if (not os.path.exists(file_folder+"drug_targets_DrugCentral.csv")):
            targets_DrugCentral = get_targets_DrugCentral(drug_names, quiet=quiet)
            targets_DrugCentral.to_csv(file_folder+"drug_targets_DrugCentral.csv")
            print(targets_DrugCentral.head())
        targets_DrugCentral = pd.read_csv(file_folder+"drug_targets_DrugCentral.csv", index_col=0)
        if (targets_DrugCentral.shape[1]<len(drug_names)):
            targets_DrugCentral = pd.concat(tuple([targets_DrugCentral]+[pd.DataFrame([],index=targets_DrugCentral.index,columns=[d]) for d in drug_names if (d not in targets_DrugCentral.columns)]), axis=1).fillna("")
        targets_list.append(targets_DrugCentral[drug_names])

    index_full = list(sorted(list(set([g for t in targets_list for g in t.index]+gene_list))))
    targets_list = [pd.concat((t, pd.DataFrame([], index=[g for g in index_full if (g not in t.index)], columns=t.columns)), axis=0).loc[index_full].fillna("") for t in targets_list]
    targets_df = reduce(lambda x,y: np.add(x,y), targets_list)
    is_inhibitor = np.vectorize(lambda x : "-" in x)
    targets_df[is_inhibitor(targets_df.values)] = "-1"
    is_activator = np.vectorize(lambda x : "+" in x)
    targets_df[is_activator(targets_df.values)] = "1"
    targets_df[targets_df==""] = "0"
    targets_df[(targets_df!="0")&(targets_df!="-1")&(targets_df!="1")] = "-1"#"1"
    targets_df = targets_df.astype(int)
    if (len(gene_list)>0):
        targets_df = targets_df.loc[gene_list]
    return targets_df

#################
##   MINERVA   ##
#################
## https://minerva.pages.uni.lu/doc/api/15.0/projects/
def get_targets_MINERVA(drug_names, quiet=False):
    '''
    Utility function using httr:GET to send queries to a given MINERVA Platform instance

    ...

    Parameters
    ----------
    drug_names : Python character string list
        list of common drug names
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    target_df : Pandas DataFrame
        rows/[lists of HGNC symbols of targets] x columns/[drug names], values are "*" (known regulator) or "" (no known regulation)
    '''
    base_url="https://minerva-dev.lcsb.uni.lu/minerva/api/projects/"
    headers = {'content-type': 'application/x-www-form-urlencoded'}
    ## 1. List all disease maps in MINERVA
    response = requests.get(base_url, headers=headers)
    assert (response.status_code==200 and (len(json.loads(response.text))>0))
    projects = json.loads(response.text)
    project_ids = [pi['projectId'] for pi in projects]
    target_lst = [[]]*len(drug_names)
    for idi, drug_name in tqdm(enumerate(drug_names)):
        for project_id in tqdm(project_ids):
            sleep(0.01)
            response = requests.get(base_url+project_id+"/drugs:search?query="+drug_name, headers=headers)
            ### 2. Omit results with zero map elements, these are drug targets not in the map
            if (response.status_code!=200):
                continue
            targets = json.loads(response.text)
            if (len(targets)==0):
                continue
            map_hits = [x for x in targets[0]["targets"] if (len(x["targetElements"])>0)]
            ### 3. Extract HGNC symbols from the 'targetParticipants' part of the results
            target_lst[idi] += [xx["resource"] for x in map_hits for xx in x["targetParticipants"]]
        target_lst[idi] = list(set(target_lst[idi]))
        if (not quiet):
            print("<MINERVA> Drug %s (%d/%d): %d targets" % (drug_name, idi+1, len(drug_names), len(target_lst[idi])))
    target_df = pd.DataFrame({dn: {x:"*" for x in target_lst[idn]} for idn, dn in enumerate(drug_names)}).fillna("")
    return target_df

#################
##  DRUGBANK   ##
#################
def get_targets_DrugBank(drug_names, path_to_drugbank, drug_fname, target_fname, quiet=False):
    '''
    Utility function which gets drug targets from a local DrugBank database

    ...

    Parameters
    ----------
    drug_names : Python character string list
        list of common drug names
    path_to_drugbank : Python character string
        (relative) path to DrugBank files
    drug_fname : Python character string
        file name of .XML file of the complete database (full_database.xml) in DrugBank
    target_fname : Python character string
        file name of .CSV file of the protein targets (all.csv) in DrugBank
    quiet : Python bool
        prints out verbose (default=False)

    Returns
    ----------
    target_df : Pandas DataFrame
        rows/[lists of HGNC symbols of targets] x columns/[drug names], values are "*" (known regulator) or "" (no known regulation)
    '''
    ## 1. Find drug names from DrugBank identifiers (using file from DrugBank after registration)
    if (not os.path.exists(path_to_drugbank+drug_fname)):
        sbcall("unzip "+"/".join(path_to_drugbank+drug_fname.split("/")[:-1])+"drugbank_all_full_database.xml.zip", shell=True)
    cmd_file="cat \'"+path_to_drugbank+drug_fname+"\'"
    cmd_drugbank_names="grep -e '^  <name>' | sed 's/  <name>//g' | sed 's/<\/name>//g'"
    cmd_drugbank_ids="grep -e '^  <drugbank-id primary=\"true\">' | sed 's/<\/drugbank-id>//g' | sed 's/<drugbank-id primary=\"true\">//g' | sed 's/ //g'"
    drugbank_ids=sbcheck_output(cmd_file+" | "+cmd_drugbank_ids, shell=True).decode("utf-8").split("\n")
    drugbank_names=sbcheck_output(cmd_file+" | "+cmd_drugbank_names, shell=True).decode("utf-8").split("\n")
    di_drugbankid2drugname = dict(zip(drugbank_ids, drugbank_names))
    ## 2. Manually retrieved from DrugBank website
    new_names = {"DB00510": "Divalproex sodium", "DB01402": "Bismuth", "DBCAT004271": 'Asparaginase', "DB00371" : "Meprobamate",
        "DB00394": "Beclomethasone dipropionate", "DB00422": "Methylphenidate", "DB00462": "Methscopolamine bromide",
        "DB00464": "Sodium tetradecyl sulfate", "DB00525": "Tolnaftate", "DB00527": "Cinchocaine", "DB00563": "Methotrexate",
        "DB05381": "Histamine", "DB00931": "Metacycline", "DB00326": "Calcium glucoheptonate",
        "DB14520": "Tetraferric tricitrate decahydrate", "DB00717": "Norethisterone", "DB01258": "Aliskiren", "DB00006" : "Bivalirudin",
    }
    di_drugbankid2drugname.update(new_names)
    ## 3. Build dictionary (keys=DrugBank IDs, values=drug_names) for the input drug names
    convert_di = {x:y for x,y in list(di_drugbankid2drugname.items()) if (y in drug_names)}
    ## 4. Get protein DB
    protein_db = pd.read_csv(path_to_drugbank+target_fname, index_col=5).query("Species=='Humans'")[["Gene Name", "Drug IDs"]]
    protein_db["Drug IDs"] = ["; ".join([convert_di[xx] for xx in list(set(x.split("; "))) if (xx in convert_di)]) for x in protein_db["Drug IDs"]]
    protein_db = protein_db.loc[protein_db["Drug IDs"]!=""]
    protein_db["presence"] = "*"
    ## 5. Decouple rows with a list of several drugs
    add_rows = protein_db.loc[np.vectorize(lambda x : "; " in x)(list(protein_db["Drug IDs"]))]
    add_rows = [[g, d, sign] for g, d_lst, sign in add_rows.values.tolist() for d in d_lst.split("; ")]
    add_rows = pd.DataFrame(add_rows, index=range(len(add_rows)), columns=protein_db.columns)
    protein_db = pd.concat((protein_db.loc[np.vectorize(lambda x : "; " not in x)(list(protein_db["Drug IDs"]))], add_rows), axis=0)
    protein_db.index = ["--".join(list(protein_db.loc[idx][["Drug IDs", "Gene Name"]])) for idx in protein_db.index]
    protein_db = protein_db.loc[~protein_db.index.duplicated()]
    target_df = protein_db.pivot(columns="Drug IDs", index="Gene Name", values="presence").fillna("")
    return target_df

####################
##  LINCS L1000   ##
####################
def get_targets_LINCS(drug_names, path_to_lincs, credentials, selection=None, nsigs=None, quiet=False):
    '''
    Utility function to retrieve drug targets from the LINCS L1000 database

    ...

    Parameters
    ----------
    drug_names : Python character string list
        list of common drug names
    path_to_lincs : Python character string
        (relative) path to LINCS files
    credentials : Python character string
        (relative) path to LINCS credentials file
    quiet : Python bool
        prints out verbose (default=False)

    Returns
    ----------
    target_df : Pandas DataFrame
        rows/[lists of HGNC symbols of targets] x columns/[drug names], values are "*" (known regulator) or "" (no known regulation)
    '''
    target_lst = [[]]*len(drug_names)
    pubchem_cids = drugname2pubchem(drug_names, lincs_args={"path_to_lincs": path_to_lincs, "credentials": credentials})
    pubchem_cids.update({"dmcm": 104999,"glycerin":753})
    user_key = get_user_key(credentials)
    for idi, drug_name in tqdm(enumerate(drug_names)):
        if (np.isnan(pubchem_cids[drug_name])):
            continue
        params = {"where": {"pubchem_cid": pubchem_cids[drug_name]}, "fields": ["target"]}
        url_perts = build_url("perts", method="filter", params=params, user_key=user_key)
        sleep(0.01)
        data = post_request(url_perts, quiet=quiet)
        if (len(data)>0):
            targets = data[0]["target"]
            target_lst[idi] = targets
        if (not quiet):
            print("<LINCS> Drug %s (%d/%d): %d targets" % (drug_name, idi+1, len(drug_names), len(target_lst[idi])))
    target_df = pd.DataFrame({dn: {x:"*" for x in target_lst[idn]} for idn, dn in enumerate(drug_names)}).fillna("")
    return target_df

##########################################
##  THERAPEUTIC TARGET DATABASE (TTD)   ##
##########################################
def get_targets_TTD(drug_names, quiet=False):
    '''
    Utility function to retrieve drug targets from the Therapeutic Target Database (TTD)

    ...

    Parameters
    ----------
    drug_name : Python character string list
        list of common drug names
    quiet : Python bool
        prints out verbose (default=False)

    Returns
    ----------
    target_df : Pandas DataFrame
        rows/[lists of HGNC symbols of targets] x columns/[drug names], values are "-" (inhibitor) or "+" (activator) or "" (no known regulation)
    '''
    ## 1. Download files (TARGET-ID and DRUG-ID matchings)
    ttd_url = "http://db.idrblab.net/ttd/sites/default/files/ttd_database/"
    target_info = "P1-01-TTD_target_download.txt"
    drug_target = "P1-07-Drug-TargetMapping.xlsx"
    sbcall("wget -nc "+ttd_url+target_info, shell=True)
    sbcall("wget -nc "+ttd_url+drug_target, shell=True)
    target_matchings = dict([x.split("\t") for x in sbcheck_output("cat "+target_info+" | grep '\tGENENAME\t' | cut -f1,3", shell=True).decode("utf-8").split("\n")[:-1]])
    drug_matchings = dict([[x.split("\t")[0],drug_names[[d.lower() for d in drug_names].index(x.split("\t")[-1].lower())]] for x in sbcheck_output("cat "+target_info+" | grep '\tDRUGINFO\t' | cut -f3,4", shell=True).decode("utf-8").split("\n")[:-1] if (x.split("\t")[-1].lower() in [d.lower() for d in drug_names])])
    ## 2. Get TARGET-DRUG MATCHINGS
    drug_targets = pd.DataFrame([x.split(",") for x in sbcheck_output("xlsx2csv "+drug_target+" | cut -d',' -f1,2,4", shell=True).decode("utf-8").split("\n")[1:-1]], columns=["Target", "Drug", "MOA"])
    drug_targets["Drug"] = [drug_matchings.get(g, "") for g in drug_targets["Drug"]]
    drug_targets = drug_targets.loc[drug_targets["Drug"]!=""]
    decrease_expr_MOA = ["inhibitor","blocker","degrader","antagonist","antisense","breaker","disrupter", "inactivator","inverse agonist","replacement","suppressor"]
    drug_targets["MOA"] = [("-" if (m in decrease_expr_MOA) else "+") for m in drug_targets["MOA"]]
    drug_targets["Target"] = [target_matchings.get(t,"") for t in drug_targets["Target"]]
    drug_targets = drug_targets.loc[drug_targets["Target"]!=""]
    ## 3. Decouple rows with several targets
    add_rows = drug_targets.loc[np.vectorize(lambda x : "; " in x)(list(drug_targets["Target"]))]
    add_rows = [[g, d, sign] for g_lst, d, sign in add_rows.values.tolist() for g in g_lst.split("; ")]
    add_rows = pd.DataFrame(add_rows, index=range(len(add_rows)), columns=drug_targets.columns)
    drug_targets = pd.concat((drug_targets.loc[np.vectorize(lambda x : "; " not in x)(list(drug_targets["Target"]))], add_rows), axis=0)
    drug_targets.index = ["--".join(list(drug_targets.loc[idx][["Drug", "Target"]])) for idx in drug_targets.index]
    drug_targets = drug_targets.loc[~drug_targets.duplicated()]
    sbcall("rm -f "+target_info+" "+drug_target, shell=True)
    target_df = drug_targets.pivot(index="Target", columns="Drug", values="MOA").fillna("")
    return target_df

##########################################
##  DrugCentral 2021                    ##
##########################################
def get_targets_DrugCentral(drug_names, quiet=False):
    '''
    Utility function to retrieve drug targets from the DrugCentral 2021

    ...

    Parameters
    ----------
    drug_name : Python character string list
        list of common drug names
    quiet : Python bool
        prints out verbose (default=False)

    Returns
    ----------
    target_df : Pandas DataFrame
        rows/[lists of HGNC symbols of targets] x columns/[drug names], values are "*" (known regulator) or "" (no known regulation)
    '''
    ffile = [[drug_names[[d.lower() for d in drug_names].index(x.split("\t")[0][1:-1].lower())], [xx for xx in x.split("\t")[1][1:-1].split("|") if ((len(xx)>1) and (xx[1].upper()==xx[1]))]] for x in sbcheck_output("wget -qO- https://unmtid-shinyapps.net/download/DrugCentral/2021_09_01/drug.target.interaction.tsv.gz | gzip -d | cut -f1,6", shell=True).decode("utf-8").split("\n")[:-1] if (x.split("\t")[0][1:-1].lower() in [d.lower() for d in drug_names])]
    ffile = [[x, y] for x,y in ffile if (len(y)>0)]
    ffile = pd.DataFrame([[drug_name, xx] for drug_name in drug_names for x in [v for u, v in ffile if (u==drug_name)] for xx in x])
    ffile = ffile.drop_duplicates(keep="first")
    ffile["presence"] = "*"
    target_df = ffile.pivot(index=1, columns=0, values="presence").fillna("")
    return target_df


#coding:utf-8

from random import seed as rseed
import os
from subprocess import call as sbcall
from subprocess import check_output as sbcheck_output
import pandas as pd
import numpy as np
from copy import deepcopy
from glob import glob

from NORDic.UTILS.DISGENET_utils import get_genes_proteins_from_DISGENET, get_user_key_DISGENET
from NORDic.UTILS.STRING_utils import get_network_from_STRING, get_app_name_STRING
from NORDic.UTILS.LINCS_utils import get_user_key
from NORDic.UTILS.utils_grn import get_genes_interactions_from_PPI, build_influences, create_grn, build_observations, infer_network, zip2df
from NORDic.UTILS.utils_grn import get_minimal_edges, general_topological_parameter, get_weakly_connected, solution2influences, save_grn, get_grfs_from_solution
from NORDic.UTILS.utils_exp import get_experimental_constraints, profiles2signatures
from NORDic.UTILS.utils_data import convert_genes_EntrezGene, convert_EntrezGene_LINCSL1000, get_all_celllines
from NORDic.UTILS.utils_plot import plot_influence_graph, plot_signatures, influences2graph

from NORDic.UTILS.utils_grn import get_grfs_from_solution
from NORDic.NORDic_NI.cytoscape_style import style_file

def network_identification(file_folder, taxon_id, path_to_genes=None, disgenet_args=None, network_fname=None, string_args=None, experiments_fname=None, lincs_args=None, edge_args=None, sig_args=None, bonesis_args=None, weights=None, seed=0, njobs=1, force_experiments=True, accept_nonRNA=False, preserve_network_sign=True):
    '''
    Generates or retrieves the optimal network model

    ...

    Parameters
    ----------
    file_folder : Python character string
        path to which files should be saved
    taxon_id : Python integer
        NCBI Taxonomy ID for the considered species
    path_to_genes : Python character string or None
        [default=None] : path to the file containing gene names (one per line)
    disgenet_args : Python dictionary or None
        [default=None] : arguments to the DisGeNET API (see test files)
    network_fname : Python character string or None
        [default=None] : path to the file containing the prior knowledge network (see test files)
    string_args : Python dictionary or None
        [default=None] : arguments to the STRING API (see test files)
    experiments_fname : Python character string or None
        [default=None] : path to the file containing the matrix of gene expression data (genes x samples) (see test files for format)
    lincs_args : Python dictionary or None
        [default=None] : arguments to the LINCS L1000 API (see test files)
    edge_args : Python dictionary or None
        [default=None] : arguments to the processing of edges (see test files)
    sig_args : Python dictionary or None
        [default=None] : arguments to the processing of signatures (see test files)
    bonesis_args : Python dictionary or None
        [default=None] : arguments to building constraints and generating solutions using BoneSIS (see test files)
    weights : Python dictionary or None
        [default=None] : weights to the optimal model selection procedure (see test files)
    seed : Python integer
        [default=0] : random seed 
    njobs : Python integer
        [default=1] : number of parallel jobs
    force_experiments : Python bool
        [default=True] : if set to True, returns an error if no experimental profile associated with the genes is found
    accept_nonRNA : Python bool
        [default=False] : if set to False, ignores gene names which cannot be converted to EntrezIDs or which are not present in LINCS L1000

    Returns
    ----------
    solution : Python character string
        optimal model selected by the procedure
    '''
    solution_fname, nsol = solution_generation(file_folder, taxon_id, path_to_genes=path_to_genes, disgenet_args=disgenet_args, network_fname=network_fname, string_args=string_args, experiments_fname=experiments_fname, lincs_args=lincs_args, edge_args=edge_args, sig_args=sig_args, bonesis_args=bonesis_args, seed=seed, njobs=njobs, force_experiments=force_experiments, accept_nonRNA=accept_nonRNA, preserve_network_sign=preserve_network_sign)

    if (nsol==0):
        return None

    print("--- SELECTION OF OPTIMAL MODEL ---")
    ###########################
    ## SELECT OPTIMAL MODEL  ##
    ###########################
    print("A. Select an optimal model based on a topological criteria... ", end=" ...")

    solutions = import_all_solutions(solution_fname)
    print("... %d solutions (%d/%d constant genes in average)" % (solutions.shape[1], np.mean([solutions.loc[(solutions[c].astype(str)=="0")|(solutions[c].astype(str)=="1")].shape[0] for c in solutions.columns]), solutions.shape[0]))
    visualize_models(solutions, file_folder)
    solution = select_optimal_model(solutions, weights, file_folder)

    return solution

def solution_generation(file_folder, taxon_id, path_to_genes=None, disgenet_args=None, network_fname=None, string_args=None, experiments_fname=None, lincs_args=None, edge_args=None, sig_args=None, bonesis_args=None, weights=None, seed=0, njobs=1, force_experiments=True, accept_nonRNA=False,preserve_network_sign=True):
    '''
    Generates or retrieves the optimal network model

    ...

    Parameters
    ----------
    file_folder : Python character string
        path to which files should be saved
    taxon_id : Python integer
        NCBI Taxonomy ID for the considered species
    path_to_genes : Python character string or None
        path to the file containing gene names (one per line)
    disgenet_args : Python dictionary or None
        arguments to the DisGeNET API (see test files)
    network_fname : Python character string or None
        path to the file containing the prior knowledge network (see test files)
    string_args : Python dictionary or None
        arguments to the STRING API (see test files)
    experiments_fname : Python character string or None
        path to the file containing the matrix of gene expression data (genes x samples) (see test files for format)
    lincs_args : Python dictionary or None
        arguments to the LINCS L1000 API (see test files)
    edge_args : Python dictionary or None
        arguments to the processing of edges (see test files)
    sig_args : Python dictionary or None
        arguments to the processing of signatures (see test files)
    bonesis_args : Python dictionary or None
        arguments to building constraints and generating solutions using BoneSIS (see test files)
    seed : Python integer
        [default=0] : random seed
    njobs : Python integer
        [default=1] : number of parallel jobs
    force_experiments : Python bool
        [default=True] : if set to True, returns an error if no experimental profile associated with the genes is found
    accept_nonRNA : Python bool
        [default=False] : if set to False, ignores gene names which cannot be converted to EntrezIDs or which are not present in LINCS L1000

    Returns
    ----------
    solution : Python character string
        optimal model selected by the procedure
    '''
    sbcall("mkdir -p "+file_folder, shell=True)
    solution_fname=file_folder+("SOLUTIONS-%d_binthres=%.3f_thresiscale=%s_score=%.2f_maxclause=%d" % (bonesis_args.get("limit", 1), sig_args.get("bin_thres", 0.5), str(lincs_args.get("thres_iscale", 0.)), string_args.get("score", 1), bonesis_args.get("max_maxclause", 5)))
    solution_fname_ls, solution_ls = glob(solution_fname+"_*.zip"), []
    #####################################
    ## RETRIEVE EXISTING RESULTS       ##
    #####################################
    if (len(solution_fname_ls)>0):
        for fname in solution_fname_ls:
            nbits = int(sbcheck_output("ls -l "+fname+"  | cut -d\" \" -f5", shell=True).decode("utf-8"))
            if (nbits==0):
                sbcall("rm "+fname, shell=True)
            else:
                solution_ls.append(fname)
    if (len(solution_ls)>=bonesis_args.get("niterations", 1)):
        print("%d solutions are already generated, and saved at %s_{1,...%d}.zip" % (len(solution_ls)*bonesis_args.get("limit", 1), solution_fname, len(solution_ls)))
        return solution_fname, len(solution_ls)

    np.random.seed(seed)
    rseed(seed)

    ## Create folder
    sbcall("mkdir -p "+file_folder, shell=True)

    print("--- DATA IMPORT ---")

    ###########################
    ## IMPORT GENE SET       ##
    ###########################
    print("1. Import gene set from %s" % ("DisGeNET" if (path_to_genes is None) else path_to_genes), end="... "),

    gene_set_file = file_folder+"GENESET.txt"
    if (path_to_genes is None and network_fname is None):
        ## Import from DisGeNET
        assert disgenet_args
        assert disgenet_args.get("credentials", False)
        assert disgenet_args.get("disease_cids", False)
        user_key = get_user_key_DISGENET(disgenet_args["credentials"])
        gene_df = get_genes_proteins_from_DISGENET(disgenet_args["disease_cids"], 
                            min_score=disgenet_args.get("min_score", 0), 
                            min_ei=disgenet_args.get("min_ei", 0),
                            min_dsi=disgenet_args.get("min_dsi", 0.25),
                            min_dpi=disgenet_args.get("min_dpi", 0),
                            user_key=user_key
        )
        model_genes = list(set([g for x in gene_df["Gene Name"] for g in x.split("; ")]))
    elif (path_to_genes is None):
        ## Import prepared network: four columns ["preferredName_A","preferred_B","sign","directed","score"], sep="\t"
        ## preferredName_A, preferredName_B are HUGO gene symbols
        ## score in [0,1], directed in {1}, sign in {1,-1} (duplicate edges which are not directed / signed)
        network = pd.read_csv(network_fname, sep="\t")
        cols = ["preferredName_A","preferredName_B","sign","directed","score"]
        assert network.shape[1] == len(cols)
        for coli, col in enumerate(network.columns):
            assert col == cols[coli]
        for col in cols:
            assert col in network.columns
        assert all([v in [1,2,-1] for v in list(network["sign"])])
        assert all([v in [1,0] for v in list(network["directed"])])
        assert all([v <= 1 and v >= 0 for v in list(network["score"])])
        model_genes = list(set(list(network["preferredName_A"])+list(network["preferredName_B"])))
    else:
        ## Import prepared gene set: one gene name per line
        assert os.path.exists(path_to_genes)
        with open(path_to_genes, "r") as f:
            model_genes = f.read().split("\n")[:-1]

    with open(gene_set_file, "w") as f:
        f.write(("\n".join(model_genes))+"\n")

    print("... %d genes imported." % len(model_genes))
    assert len(model_genes)>0

    ###########################
    ## IMPORT NETWORK        ##
    ###########################
    print("2. Import network from %s" % ("STRING" if (network_fname is None) else network_fname), end="... ")

    network_file = file_folder+"NETWORK.tsv"
    if (not os.path.exists(network_file)):
        if (network_fname is None):
            ## Import from STRING
            app_name = get_app_name_STRING(string_args["credentials"])
            network = get_network_from_STRING(model_genes, 
                            taxon_id,
                            network_type=string_args.get("network_type", "functional"),
                            min_score=string_args.get("score", 0), 
                            app_name=app_name,
                            quiet=True
            )
        else:
            ## Import prepared network: four columns ["preferredName_A","preferred_B","sign","directed","score"], sep="\t"
            ## preferredName_A, preferredName_B are HUGO gene symbols
            ## score in [0,1], directed in {1,0}, sign in {1,2,-1}
            network = pd.read_csv(network_fname, sep="\t")
            cols = ["preferredName_A","preferredName_B","sign","directed","score"]
            assert network.shape[1] == len(cols)
            for coli, col in enumerate(network.columns):
                assert col == cols[coli]
            for col in cols:
                assert col in network.columns
            assert all([v in [1,2,-1] for v in list(network["sign"])])
            assert all([v in [1,0] for v in list(network["directed"])])
            assert all([v <= 1 and v >= 0 for v in list(network["score"])])

        network.to_csv(network_file, index=None, sep="\t")
    network = pd.read_csv(network_file, sep="\t")
    network_genes = list(set([g for a in ["preferredName_A","preferredName_B"] for g in list(network[a])]))

    print("... %d edges in model (including %d directed edges) with a total of %d non-isolated genes" % (network.shape[0], network[network["directed"]==1].shape[0], len(network_genes)))

    ## Write down the list of missing genes in PPI
    if (len(model_genes)>len(network_genes)):
        with open(file_folder+"missing_genes_in_PPI.txt", "w") as f:
            missing_genes = [g for g in model_genes if (g not in network_genes)]
            f.write("\n".join(missing_genes))

    if (network.shape[0]==0):
        raise ValueError("No network could be found.")
    assert network.shape[0]>0
    model_genes = network_genes ## trim out isolated nodes

    ###########################
    ## IMPORT EXPERIMENTS    ##
    ###########################
    print("3. Import experiments from %s" % ("LINCS L1000" if (experiments_fname is None) else experiments_fname), end="... "),

    ## Get cell line names from LINCS L1000
    if (len(lincs_args.get("cell_lines", []))==0):
        user_key = get_user_key(lincs_args["credentials"])
        cell_lines = get_all_celllines(pert_inames, user_key)
    else:
        cell_lines = lincs_args["cell_lines"]

    print("\n\t%d cell lines are considered (%s)" % (len(cell_lines), cell_lines))
    if (force_experiments and (experiments_fname is None)):
        assert len(cell_lines)>0

    pert_df = None
    if (experiments_fname is None):
        ## Convert gene names in EntrezGene IDs
        entrezgene_fname=file_folder+"ENTREZGENES.csv"
        if (not os.path.exists(entrezgene_fname)):
            app_name = get_app_name_STRING(string_args["credentials"])
            probes = convert_genes_EntrezGene(model_genes, taxon_id, app_name=app_name)
            probes.to_csv(entrezgene_fname)
        probes = pd.read_csv(entrezgene_fname,index_col=0)

        if (probes.shape[0]>probes[probes["Gene ID"]!="-"].shape[0]):
            with open(file_folder+"missing_genes_in_EntrezID.txt", "w") as f:
                missing_genes = [g for g in probes.index if (g not in list(probes[probes["Gene ID"]!="-"].index))]
                f.write("\n".join(missing_genes))

        if (probes[probes["Gene ID"]=="-"].shape[0]>0 and not accept_nonRNA):
            print("\tNot found genes: %s" % str(list(probes[probes["Gene ID"]=="-"].index)))
        probes = probes[probes["Gene ID"]!="-"]
        print("\t%d genes available (convertable to EntrezIDs)" % probes.shape[0])
        if (probes.shape[0]<len(model_genes) and not accept_nonRNA):
            print("\t\tCheck that\n\t\t1. All input genes are HUGO Gene Symbols / Ensembl IDs / HGNC IDs / STRING IDs;\n\t\t2. The correct taxon id (%d) was provided." % taxon_id)

        ## Convert EntrezGene IDs to LINCS L1000 symbols
        symbols_fname=file_folder+"SYMBOLS.csv"
        if (not os.path.exists(symbols_fname)):
            user_key = get_user_key(lincs_args["credentials"])
            pert_df = convert_EntrezGene_LINCSL1000(file_folder, list(probes["Gene ID"]), user_key)
            pert_df.to_csv(symbols_fname)
        pert_df = pd.read_csv(symbols_fname, index_col=0)

        missing_genes = [list(probes.index)[list(probes["Gene ID"]).index(g)] for g in pert_df.index if (g not in pert_df[pert_df["Gene Symbol"]!="-"].index)]
        pert_df = pert_df[pert_df["Gene Symbol"]!="-"]
        pert_inames = list(pert_df["Gene Symbol"])
        pert_df.index = pert_inames
        entrez_ids = list(pert_df["Entrez ID"])
        print("\t\t%d/%d genes retrieved in LINCS L1000" % (pert_df.shape[0], probes.shape[0]))

        if (len(missing_genes)>0):
            with open(file_folder+"missing_genes_in_LINCSL1000.txt", "w") as f:
                f.write("\n".join(missing_genes))

    add_rows_profiles = ["annotation", "perturbed", "perturbation", "cell_line", "sigid", "interference_scale"]
    thres_iscale = lincs_args.get("thres_iscale", None)
    profiles_fname = file_folder+"PROFILES_"+"-".join(cell_lines[:4])+"_thres=%s.csv" % str(thres_iscale)
    if (not os.path.exists(profiles_fname)):
        if (experiments_fname is None):
            full_profiles_fname = file_folder+"FULL_PROFILES_"+"-".join(cell_lines[:4])+".csv"
            assert string_args
            assert string_args.get("credentials", False)
            assert lincs_args
            assert lincs_args.get("credentials", False)

            ## Get experimental profiles
            pert_types = lincs_args.get("pert_types", ["trt_sh","trt_oe","trt_xpr"])
            selection = lincs_args.get("selection", "distil_ss")
            path_to_lincs = lincs_args.get("path_to_lincs", "./")
            nsigs = lincs_args.get("nsigs", 2)
            user_key = get_user_key(lincs_args["credentials"])
            pert_di = pert_df[["Entrez ID","Gene Symbol"]].to_dict()["Entrez ID"]
            if (not os.path.exists(full_profiles_fname)):
                full_profiles = get_experimental_constraints(file_folder, cell_lines, pert_types, pert_di, taxon_id, selection, user_key, path_to_lincs, thres_iscale=thres_iscale, nsigs=nsigs, quiet=False)
                full_profiles.to_csv(full_profiles_fname)
            else:
                full_profiles = pd.read_csv(full_profiles_fname, index_col=0, low_memory=False)

            if (full_profiles.shape[1]==0):
                profiles = None
            else:
                if (thres_iscale is not None):
                    profiles_ = full_profiles[[c for ic, c in enumerate(full_profiles.columns) if (float(list(full_profiles.loc["interference_scale"])[ic])>thres_iscale)]]
                    if (profiles_.shape[1]==0):
                        raise ValueError("The selected value of thres_iscale (%s) is too large, maximum value in collected experiments=%s" % (thres_iscale, np.max(full_profiles.loc["interference_scale"].values)))
                    min_iscale = np.min(profiles_.loc["interference_scale"].values)
                else:
                    profiles_ = full_profiles
                    min_iscale = None
                profiles = profiles_.loc[[i for i in profiles_.index if (i != "interference_scale")]]

                Nfullprofs = len(list(set(["_".join(full_profiles[c].loc[["perturbed", "perturbation", "cell_line"]]) for c in full_profiles.columns])))
                Nprofs = len(list(set(["_".join(profiles[c].loc[["perturbed", "perturbation", "cell_line"]]) for c in profiles.columns])))
                print("\t\t%d unique experiments (including %d with criterion thres_iscale > %s, min_value %s)" % (Nfullprofs, Nprofs, thres_iscale, min_iscale))
                profiles = profiles.loc[~profiles.index.duplicated()]
                profiles.index = [g if (g in add_rows_profiles) else pert_df.loc[pert_df["Entrez ID"]==float(g)]["Gene Symbol"].values[0] for g in list(profiles.index)]

            if (profiles is None):
                profiles = pd.DataFrame([], index=pert_inames, columns=["0","1"]).astype(str)
                for a in add_rows_profiles:
                    profiles.loc[a] = ["nan"]*profiles.shape[1]

        else:

            profiles = pd.read_csv(experiments_fname, sep=",", index_col=0, na_filter=False)
            ## Pandas DataFrame with index=HUGO gene symbols, columns=sample names, and 5 additional rows
            ## 'annotation' 2 if the sample is treated, 1 otherwise
            ## 'perturbed' gene which is perturbed
            ## 'perturbation' experiment UNIQUE identifier, common to samples from the same experiment
            ## 'cell_line' experiment UNIQUE identifier, common to samples from the same experiment
            ## 'sigid' sample unique identifier from LINCS L1000 (can be filled with NaNs)
            ## /!\ SHOULD ALREADY BE BINARIZED
            assert "annotation" in list(profiles.index) and all([v in ["1", "2"] for v in list(profiles.loc["annotation"])])
            assert "perturbed" in list(profiles.index)
            assert "perturbation" in list(profiles.index) and all([v in ["KD", "OE", "None"] for v in list(profiles.loc["perturbation"])])
            assert "cell_line" in list(profiles.index)
            assert "sigid" in list(profiles.index)
            assert all([v in ["0","1",np.nan] for v in list(profiles.loc[[g for g in model_genes if (g in profiles.index)]].values.flatten())])
            profiles = profiles[[c for c in profiles.columns if ((profiles.loc["perturbed"][c] in model_genes+["None"]) and (profiles.loc["cell_line"][c] in cell_lines))]]
            profiles = profiles.loc[[g for g in list(profiles.index) if (g in model_genes+add_rows_profiles)]]
        profiles.to_csv(profiles_fname)

    profiles = pd.read_csv(profiles_fname, index_col=0, low_memory=False)
    if (not pd.isnull(profiles.values).all()):
        nconds = len(set(["_".join([profiles.loc[v][c] for v in ["perturbed","perturbation","cell_line"]]) for c in profiles.columns]))
    else:
        nconds = 0
    ngenes = len([p for p in profiles.index if (p not in add_rows_profiles)])
    print("... %d genes in %d profiles (%d experiments)" % (0 if (profiles is None) else ngenes, 0 if (profiles is None) else profiles.shape[1], 0 if (profiles is None) else nconds))
    if (profiles is None and force_experiments):
        raise ValueError("No experimental profile could be found.")
        assert (not force_experiments) or ((profiles is not None) and profiles.shape[1]>0)
    assert (not force_experiments) or ((profiles is not None) and profiles.shape[0]-len(add_rows_profiles)>0)

    ## Write down the list of missing genes in profiles
    if ((profiles is not None) and (pert_df is not None) and (pert_df.shape[0]>ngenes)):
        with open(file_folder+"missing_genes_in_profiles.txt", "w") as f:
            missing_genes = [g for g in pert_inames if (g not in list(profiles.index))]
            f.write("\n".join(missing_genes))

    print("\n--- CONSTRAINT BUILDING ---")

    #####################################
    ## FILTERING OUT EDGES             ##
    #####################################
    print("1. Filtering out edges by minimum set of edges with highest score which preserve connectivity...", end=" ")

    network = pd.read_csv(network_file, sep="\t")

    ## Filter out edges which do not involve nodes which are both present in the experiments
    if (not accept_nonRNA and profiles is not None):
        test_in_profiles = np.vectorize(lambda x : x in [p for p in list(profiles.index) if (p not in add_rows_profiles)])
        network = network.loc[(test_in_profiles(network["preferredName_A"])&test_in_profiles(network["preferredName_B"]))]
        ## Write down the list of missing genes in profiles
        genes_in_network = list(set(list(network["preferredName_A"])+list(network["preferredName_B"])))
        if (len(model_genes)>len(genes_in_network)):
            with open(file_folder+"missing_genes_in_filtered_network.txt", "w") as f:
                missing_genes = [g for g in list(profiles.index) if (g not in genes_in_network+add_rows_profiles)]
                f.write("\n".join(missing_genes))
        model_genes = genes_in_network
    ## Ensure that all considered edges are unique
    network = network.drop_duplicates(keep="first")
    print("... %d unique edges involving genes both in experiments (%d genes in total)" % (network.shape[0], len(model_genes)))
    plot_influence_graph(network, "preferredName_A", "preferredName_B", "sign", "directed", file_folder+"network", True)

    score_thres = 0 if (string_args is None) else string_args.get("score", 0)
    edges_file = file_folder+"EDGES_score=%f.tsv" % (score_thres)
    if (not os.path.exists(edges_file) and ((network["sign"]==2).any() or (network["directed"]==0).any())):
        network_df = get_genes_interactions_from_PPI(network, connected=(edge_args is not None and edge_args.get("connected", True)), score=score_thres, filtering=(edge_args is not None and edge_args.get("filter", True)))
        network_df.to_csv(edges_file, sep="\t", index=None)
    elif (not os.path.exists(edges_file)):
        network_df = deepcopy(network)
        network_df["sscore"] = np.multiply(network_df["sign"], network_df["score"])
        network_df = network_df[["preferredName_A","preferredName_B", "sscore"]]
        network_df.columns = ["Input", "Output", "SSign"]
        network_df.to_csv(edges_file, sep="\t", index=None)
    network_df = pd.read_csv(edges_file, sep="\t")
    ## Restrict genes to non-isolated genes
    model_genes = [m for m in list(set(list(network_df["Input"])+list(network_df["Output"]))) if ((profiles is None) or (m in model_genes))]

    print("... score_STRING %f\t%d genes (non isolated in PPI)\t%d edges in PPI" % (string_args["score"], len(model_genes), network_df.shape[0]))

    ## Restrict profiles to non isolated genes 
    if (profiles is not None):
        profiles = profiles.loc[[m for m in model_genes if (m in profiles.index)]+[a for a in add_rows_profiles if (a in profiles.index)]]

    ## Write down the list of missing genes in network
    ngenes = len([p for p in profiles.index if (p not in add_rows_profiles)])
    if ((profiles is not None) and (len(model_genes)>ngenes)):
        with open(file_folder+"missing_genes_in_network.txt", "w") as f:
            missing_genes = [g for g in model_genes if (g not in list(profiles.index))]
            f.write("\n".join(missing_genes))

    ###################################
    ## BUILD TOPOLOGICAL CONSTRAINTS ##
    ###################################
    print("2. Build topological constraints from filtered edges using gene expression data... ", end=" ...")

    influences_fname = file_folder+"INFLUENCES_"+"-".join(cell_lines[:4])+"_tau="+str(edge_args.get("tau", 0))+"_beta="+str(edge_args.get("beta", 1))+"_score_thres="+str(score_thres)+".csv"
    expr_df = None if (profiles is None) else profiles.loc[[g for g in profiles.index if (g not in add_rows_profiles)]].apply(pd.to_numeric)
    expr_df=expr_df if (not pd.isnull(expr_df.values).all()) else None
    if (not os.path.exists(influences_fname) and (network_fname is None or edge_args.get("tau", False))):
        influences = build_influences(network_df, edge_args.get("tau", 0), beta=edge_args.get("beta", 1), cor_method=edge_args.get("cor_method", "pearson"), expr_df=expr_df, accept_nonRNA=accept_nonRNA)
        influences.to_csv(influences_fname)
    elif (not os.path.exists(influences_fname)):
        if (preserve_network_sign or (expr_df is None)):
            ## use the sign given by the score in the existing network
            network_df.index = ["-".join(list(map(str, list(network_df.loc[idx][["Input","Output"]])))) for idx in network_df.index]
            influences = network_df.groupby(level=0).max()
            network_df.index = range(network_df.shape[0])
            influences = influences.pivot_table(index="Input",columns="Output",values="SSign", fill_value=0)
            missing_index = [g for g in influences.columns if (g not in influences.index)]
            if (len(missing_index)>0):
                missing_index_df = pd.DataFrame([], index=missing_index, columns=influences.columns).fillna(0)
                influences = pd.concat((influences, missing_index_df), axis=0)
            missing_columns = [g for g in influences.index if (g not in influences.columns)]
            if (len(missing_columns)>0):
                missing_columns_df = pd.DataFrame([], index=influences.index, columns=missing_columns).fillna(0)
                influences = pd.concat((influences, missing_columns_df), axis=1)
            assert influences.shape[0]==influences.shape[1]
            influences[influences<0] = -1
            influences[influences>0] = 1
        else:
            ## compute the sign based on the gene pairwise correlation on expression data
            influences = build_influences(network_df, 0, beta=edge_args.get("beta", 1), cor_method=edge_args.get("cor_method", "pearson"), expr_df=expr_df, accept_nonRNA=accept_nonRNA)
        influences.to_csv(influences_fname)
    influences = pd.read_csv(influences_fname, index_col=0, header=0)
    if (not pd.isnull(profiles.values).all() is not None):
        profiles = profiles.loc[[m for m in list(influences.index) if (m in profiles.index)]+[a for a in add_rows_profiles if (a in profiles.index)]]
    print("... %d negative, %d positive undirected interactions (%d edges in total), %d non isolated genes in experiments" % (np.sum(influences.values==-1), np.sum(influences.values==1), np.sum(influences.abs().values), influences.shape[0]))
    assert np.sum(influences.abs().values) > 0
    
    influences_df = influences.melt(ignore_index=False)
    influences_df["id"] = influences_df.index
    influences_df = influences_df[influences_df["value"]!=0]
    plot_influence_graph(influences_df, "id", "variable", "value", None, influences_fname.split(".csv")[0], optional=True)

    ###################################
    ## BUILD DYNAMICAL CONSTRAINTS   ##
    ###################################
    print("3. Build dynamical constraints by binarization of experimental profiles... ", end=" ...")

    ## Signatures are vectors containing values {NaN,0,1}
    sigs_fname = file_folder+"SIGNATURES_"+"-".join(cell_lines[:4])+"_binthres="+str(sig_args.get("bin_thres", 0.5))+"_thres_iscale="+str(lincs_args.get("thres_iscale", None))+".csv"
    if (not pd.isnull(profiles.values).all() and not os.path.exists(sigs_fname)):
        save_fname=file_folder+"CELL"
        if (experiments_fname is None):
            assert lincs_args
            assert lincs_args.get("credentials", False)
            assert lincs_args.get("path_to_lincs", False)
            save_fname=file_folder+"CELL"
            user_key = get_user_key(lincs_args["credentials"])
            profiles_index = list(profiles.index)
            profiles.index = [int(pert_df.loc[g]["Entrez ID"]) if (g not in add_rows_profiles) else g for g in profiles_index]
            signatures = profiles2signatures(profiles, user_key, lincs_args["path_to_lincs"], save_fname, thres=sig_args.get("bin_thres", 0.5), selection=lincs_args.get("selection", "distil_ss"), backgroundfile=True, bin_method=sig_args.get("bin_method", "binary"), nbackground_limits=(sig_args.get("min_selection", 4), sig_args.get("max_selection", 50)))
            profiles.index = profiles_index
            signatures.index = [pert_inames[list(pert_df["Entrez ID"]).index(e)] for e in signatures.index]
        else:
            conditions = ["_".join([profiles.loc[v][c] for v in ["perturbed","perturbation","cell_line"]]) if (profiles.loc["annotation"][c]=="2") else "initial_"+profiles.loc["cell_line"][c] for c in profiles.columns]
            signatures = profiles.loc[[g for g in profiles.index if (g not in add_rows_profiles)]].apply(pd.to_numeric)
            signatures.columns = conditions
        signatures.to_csv(sigs_fname)
    elif (pd.isnull(profiles.values).all()):
        signatures = pd.DataFrame([], index=model_genes, columns=["0","1"])
        nb_absent_genes = len(model_genes)
        nb_constant_genes = 0
        nb_undetermined_genes, nexps, ncells, ngenes = [0]*4
        fnorm = np.nan
    else:
        signatures = pd.read_csv(sigs_fname, index_col=0, header=0).dropna(how="all")

    if (not pd.isnull(signatures.values).all()):
        signatures_copy = deepcopy(signatures)
        signatures_copy[signatures_copy==0] = -1
        signatures_copy = signatures_copy.fillna(0)
        fnorm = np.linalg.norm(signatures_copy.values)

        nb_absent_genes = len(model_genes)-signatures.shape[0]
        nb_constant_genes = signatures.loc[(signatures.mean(axis=1, skipna=True)==0)|(signatures.mean(axis=1, skipna=True)==1)].shape[0]+nb_absent_genes
        nb_undetermined_genes = signatures.loc[pd.isnull(signatures.mean(axis=1, skipna=True))].shape[0]+nb_absent_genes
        nexps, ncells, ngenes = len(set([c for c in signatures.columns if ("initial" not in c)])), len(set([c for c in signatures.columns if ("initial" in c)])), signatures.shape[0]

        plot_signatures(signatures, perturbed_genes=list(set(list(profiles.loc["perturbed"]))), fname=file_folder+"signatures_binthres="+str(sig_args.get("bin_thres",0.5))+"_thresiscale="+str(lincs_args.get("thres_iscale",0.)))

        print(("... %d experiments on %d cell lines and %d/%d genes (Frobenius norm signature matrix: %.3f, %d possibly constant genes: %.1f" % (nexps, ncells, ngenes, len(model_genes), fnorm, nb_constant_genes, nb_constant_genes*100/len(model_genes)))+"%, "+str(nb_undetermined_genes)+" genes with undetermined status")
    else:
        print("... no experiments.")
    print("\n--- INFER BOOLEAN NETWORK ---")

    ###########################
    ## INFER BOOLEAN NETWORK ##
    ###########################
    print("1. Generate solutions from topological & dynamical constraints... ", end=" ...")

    solution_fname_ls = glob(solution_fname+"_*.zip")
    solution_ls = []
    if (len(solution_ls)<bonesis_args.get("niterations", 1)):
        grn = create_grn(influences, exact=bonesis_args.get("exact", False), max_maxclause=bonesis_args.get("max_maxclause", 3))
        gene_list = grn.nodes
        if (not pd.isnull(signatures.values).all()):
            signatures = signatures[[c for c in signatures.columns if (("initial" in c) or any([(g in c) for g in gene_list]))]]
            signatures = signatures[list(sorted(list(signatures.columns)))]
            cols = [c for c in signatures.columns if ("initial" not in c)]
        if (pd.isnull(signatures.values).all() or sig_args.get("bin_thres",0.5)==0):
            signatures = []
        ## If considering only a subset of experiments
        elif (len(bonesis_args.get("exp_ids", []))>0):
            exps = [cols[i] for i in bonesis_args["exp_ids"]]
            signatures = signatures[[e for e in exps]]+signatures[["initial_"+cell for cell in list(set([c.split("_")[-1] for c in cols]))]]
        BO = build_observations(grn, signatures)
        ## File-readable by BoneSiS: to easily modify constraints
        BO.boolean_networks().standalone(output_filename=solution_fname+".asp")
        nsolutions = infer_network(BO, fname=solution_fname, limit=bonesis_args.get("limit", 1), use_diverse=bonesis_args.get("use_diverse", True), niterations=bonesis_args.get("niterations",1), njobs=njobs)
        if (sum(nsolutions)==0):
            print("No solution found. Try decreasing value bin_thres=%.3f in [0,0.5], or decreasing STRING score=%.2f in [0,1], or increasing max_maxclause=%d, or increasing thres_iscale=%s." % (sig_args.get("bin_thres",0.5),string_args.get("score", 1), bonesis_args.get("max_maxclause", 5), lincs_args.get("thres_iscale", None)))
            sbcall("rm -f "+solution_fname+"*.zip", shell=True)
            return None, 0
        else:
            params = {"file_folder":file_folder, "taxon_id":taxon_id, "path_to_genes":path_to_genes, 
              "disgenet_args": disgenet_args, "string_args" : string_args, "lincs_args": lincs_args, 
              "edge_args" : edge_args, "sig_args" : sig_args, "bonesis_args": bonesis_args, 
              "weights": weights, "seed": seed, "network_fname" : network_fname, "njobs": njobs}
            import json
            with open(solution_fname+'.json', 'w', encoding='utf-8') as f:
                json.dump(params, f, ensure_ascii=False, indent=4)
            return solution_fname, nsolutions

def import_all_solutions(solution_fname, quiet=False):
    '''
    Import all solutions which have been generated

    ...

    Parameters
    ----------
    solution_fname : Python character string
        header of solution files
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    solutions : Pandas DataFrame
        rows/[genes] x columns/[solution IDs] containing regulatory functions for each gene in each solution
    '''
    fname_ls = glob(solution_fname+"_*.zip")
    assert len(fname_ls)>0
    fname_ls = list(sorted(fname_ls, key=lambda x: int(x.split(".zip")[0].split("_")[-1])))
    sol_ls, nsol = [], 1
    try:
        for fi, fname in enumerate(list(sorted(fname_ls))):
            solutions = zip2df(fname)
            cols = list(range(nsol,nsol+solutions.shape[0]))
            nsol = np.max(cols)+1
            sol_ls.append(pd.DataFrame(solutions.T.values, index=solutions.columns, columns=cols))
    except:
        raise ValueError("'"+fname+"' not loaded.")
    nsol -= 1
    if (len(sol_ls)>1):
        solutions = sol_ls[0].join(sol_ls[1:], how="outer")
    else:
        solutions = sol_ls[0]
    if (not quiet):
        print("%d solutions (%d unique solutions)" % (solutions.shape[1], solutions.T.drop_duplicates().shape[0]))
    return solutions

def visualize_models(sols, file_folder):
    '''
    Selection of an optimal model in a set of solutions, based on a topology-based desirability function

    ...

    Parameters
    ----------
    sols : Pandas DataFrame
        rows/[genes] x columns/[solution IDs]
    weights : Python dictionary
        weight for each graph characteristic

    Returns
    ----------
    solution : Pandas DataFrame
        rows/[genes] x column/[solution ID] selected solution
    '''
    ## * Minimal model (number of edges) *
    minimal, nminimal = get_minimal_edges(sols)
    influences_minimal = solution2influences(minimal)
    influences2graph(influences_minimal, fname=file_folder+"inferred_minimal_solution", optional=False)
    print("<MODEL VISUALIZATION> Minimal solution: %d edges" % nminimal)
    ## * Maximal model (number of edges) *
    maximal, nmaximal = get_minimal_edges(sols, maximal=True)
    influences_maximal = solution2influences(maximal)
    influences2graph(influences_maximal, fname=file_folder+"inferred_maximal_solution", optional=False)
    print("<MODEL VISUALIZATION> Maximal solution: %d edges" % nmaximal)
    return None

def select_optimal_model(sols, weights, file_folder):
    ## * Maximizer of general topological criterion (GTP)
    GTPs = [general_topological_parameter(solution2influences(sols[c]), weights) for c in sols.columns]
    GTP_df = pd.DataFrame([GTPs], index=["GTP"], columns=sols.columns)
    print(GTP_df.head())
    GTP_df.to_csv(file_folder+"GPT.csv")
    max_criterion_solution = sols[sols.columns[np.argmax(GTPs)]]
    max_criterion_influences = solution2influences(max_criterion_solution)
    influences2graph(max_criterion_influences, file_folder+"inferred_max_criterion_solution", optional=False)
    print("<MODEL SELECTION> Saving optimal model in '%s/solution.bnet'" % file_folder, end =" ...")
    save_grn(max_criterion_solution, file_folder+"solution")
    print("... saved!")
    return max_criterion_solution

def solution2cytoscape(solution, fname):
    '''
    Convert a solution into a Cytoscape-readable file

    ...

    Parameters
    ----------
    solution : Pandas Series
        rows/[genes]
    fname : Python character string
        path to Cytoscape-readable SIF and XML (style) files (no extension)

    Returns
    ----------
    None
        writes a Cytoscape-readable file
    '''
    ## 1. Create SIF file
    target = []
    grfs = get_grfs_from_solution(solution)
    target = [" ".join([r, "->"+("+" if (grfs[g][r]>0) else "-"), g]) for g in grfs for r in grfs[g]]
    with open(fname+".sif", "w+") as f:
        f.write("\n".join(target))

    ## 2. Build associated Cytoscape style file
    target = []
    target += style_file[0]
    target.append("<visualProperty default=\""+fname+"\" name=\"SOLUTION\"/>")
    target += style_file[1]
    target += ["<visualProperty default=\"#FFCC99\" name=\"NODE_FILL_COLOR\">", "<discreteMapping attributeName=\"name\" attributeType=\"string\">"]
    target += ["</discreteMapping>", "</visualProperty>"]
    target += style_file[2]

    with open(fname+".xml", "w+") as f:
        f.write("\n".join(target))



#coding: utf-8

from time import sleep
import os
import requests
import pandas as pd
import numpy as np
from io import StringIO
from copy import deepcopy
from subprocess import check_output as sbcheck_output
from subprocess import call as sbcall

string_api_url = lambda v : "https://version-"+"-".join(v.split("."))+".string-db.org/api"

def get_app_name_STRING(fname):
    '''
    Retrieves app name from STRING to interact with the API

    ...

    Parameters
    ----------
    fname : Python character string
        path to file with a unique line = email adress

    Returns
    ----------
    app_name : Python character string
        identifier for the STRING API
    '''
    with open(fname, "r") as f:
        app_name = f.read().split("\n")[0]
    return app_name 

def get_protein_names_from_STRING(gene_list, taxon_id, app_name=None, version="11.5", quiet=False):
    '''
    Retrieves protein IDs in STRING associated with input genes in the correct species

    ...

    Parameters
    ----------
    genes_list : Python character list
        list of gene symbols
    taxon_id : Python integer
        taxon ID from NCBI
    version : Python character string
        [default="11.5"] : STRING version
    app_name : Python character string
        [default=None] : identifier to access STRING
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    res_df : Pandas DataFrame
        rows/[row number] x columns/["queryItem", "stringId", "preferredName", "annotation"]
    '''
    assert app_name
    assert taxon_id
    if (not quiet):
        print("<STRING> Getting the STRING name mapping for genes")
    output_format = "tsv"
    method = "get_string_ids"
    params = {
        "identifiers" : "\r".join(gene_list), # your protein list
        "species" : taxon_id, # species NCBI identifier
        "echo_query": 1,
        "limit" : 1, # only one (best) identifier per input protein
        "echo_query" : 1, # see your input identifiers in the output
        "caller_identity" : app_name # your app name
    }
    request_url = "/".join([string_api_url(version), output_format, method])
    results = requests.post(request_url, data=params).text
    sleep(1)
    from io import StringIO
    res_df = pd.read_csv(StringIO(results), sep="\t")
    if ("Error" in res_df.columns):
        print("<STRING_utils> Error from STRING: %s" % str(res_df["ErrorMessage"]))
        return None
    return res_df[["queryItem", "stringId", "preferredName", "annotation"]]

def get_image_from_STRING(my_genes, taxon_id, file_name="network.png", min_score=0, network_flavor="evidence", network_type="functional", app_name=None, version="11.5", quiet=False):
    '''
    Retrieves protein IDs in STRING associated with input genes in the correct species

    ...

    Parameters
    ----------
    genes_list : Python character list
        list of gene symbols
    taxon_id : Python integer
        taxon ID from NCBI
    file_name : Python character string
        [default="network.png"] : image file name
    min_score : Python float
        [default=0] : confidence lower threshold (in [0,1])
    network_flavor : Python character string
        [default="evidence"] : show links related to ["confidence", "action", "evidence"]
    network_type : Python character string
        [default="functional"] : show "functional" or "physical" network
    app_name : Python character string
        [default=None] : identifier for STRING
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    None
        writes the network image to a file file_name
    '''
    assert app_name
    assert taxon_id
    assert min_score>=0 and min_score<=1
    assert network_flavor in ["confidence", "action", "evidence"]
    assert network_type in ["functional", "physical"]
    if (not quiet):
        print("<STRING> Getting an image of the STRING network")
    output_format = "highres_image"
    method = "network"
    request_url = "/".join([string_api_url(version), output_format, method])
    params = {"identifiers" : "%0d".join(my_genes), 
        "species" : taxon_id,
        "add_white_nodes": 15, # add 15 white nodes to my protein 
        "network_flavor": network_flavor, 
        "network_type": network_type, 
        "required_score": int(min_score*1000),
        "hide_disconnected_nodes": 1,
        "hide_node_labels": 0,
        "show_query_node_labels": 0,
        "caller_identity" : app_name, 
    }
    response = requests.post(request_url, data=params)
    if (not quiet):
        print("Saving interaction network to %s" % file_name)
    with open(file_name, 'wb') as fh:
        fh.write(response.content)
    sleep(1)

def get_network_from_STRING(gene_list, taxon_id, min_score=0, network_type="functional", add_nodes=0, app_name=None, version="11.5", quiet=False):
    '''
    Retrieves undirected and unsigned interactions from the STRING database

    ...

    Parameters
    ----------
    gene_list : Python character string list
        list of gene symbols
    taxon_id : Python integer
        NCBI taxonomy ID
    min_score : Python integer
        [default=0] : minimum STRING combined edge score in [0,1]
    network_type : Python character string
        [default="functional"] : returns "functional" or "physical" network
    add_nodes : Python integer
        [default=0] : add nodes *in the closest interaction neighborhood* involved with the genes in @gene_list if set to 1
    app_name : Python character string
        [default=None] : identifier for STRING
    version : Python character string
        [default="11.5"] : STRING version
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    network : Pandas DataFrame
        rows/[row number] x columns/["preferredName_A","preferredName_B","score","directed"]
    '''
    assert app_name
    assert taxon_id
    assert min_score >= 0 and min_score <= 1
    results = get_protein_names_from_STRING(gene_list, taxon_id, app_name=app_name, version=version, quiet=quiet)
    id_di = {}
    my_genes = []
    for line in range(len(results.index)):
        input_identifier, string_identifier = results["queryItem"][line], results["stringId"][line]
        my_genes.append(string_identifier)
        id_di.setdefault(string_identifier, input_identifier)
    if (not quiet):
        print("<STRING> Getting the STRING network interactions")
    output_format = "tsv"
    method = "network"
    request_url = "/".join([string_api_url(version), output_format, method])
    params = {
        "identifiers" : "%0d".join(my_genes), # your protein
        "species" : taxon_id, # species NCBI identifier 
        "required_score" : int(min_score*1000), # in 0 - 1000, 0 : get all edges
        "network_type": network_type,
        "add_nodes": add_nodes,
        "show_query_node_labels": 0,
        "caller_identity" : app_name # your app name
    }
    if (not quiet):
        print("<STRING> Getting the STRING network interactions")
    response = requests.post(request_url, data=params).text
    sleep(1)
    network = pd.read_csv(StringIO(response), sep="\t")
    network["preferredName_A"] = [id_di.get(x, x) for x in list(network["preferredName_A"])]
    network["preferredName_B"] = [id_di.get(x, x) for x in list(network["preferredName_B"])]
    network = network[["preferredName_A","preferredName_B","score"]]
    network["sign"] = [2]*network.shape[0]
    network["directed"] = [0]*network.shape[0]
    network = network[["preferredName_A","preferredName_B","sign","directed","score"]]
    network = network.drop_duplicates(keep="first")
    network.sort_values(by="score", ascending=False)
    return network

def get_interactions_partners_from_STRING(gene_list, taxon_id, min_score=0, network_type="functional", add_nodes=0, limit=5, app_name=None, version="11.5", quiet=False):
    '''
    Retrieves undirected and unsigned interactions from the STRING database

    ...

    Parameters
    ----------
    gene_list : Python character string list
        list of gene symbols
    taxon_id : Python integer
        NCBI taxonomy ID
    min_score : Python integer
        [default=0] : minimum STRING combined edge score in [0,1]
    network_type : Python character string
        [default="functional"] : returns "functional" or "physical" network
    limit : Python integer
        [default=5] : limits the number of interaction partners retrieved per protein (most confident interactions come first)
    app_name : Python character string
        [default=None] : identifier for STRING
    version : Python character string
        [default="11.5"] : STRING version
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    network : Pandas DataFrame
        rows/[row number] x columns/["preferredName_A","preferredName_B","score","directed"]
    '''
    assert app_name
    assert taxon_id
    assert min_score >= 0 and min_score <= 1
    results = get_protein_names_from_STRING(gene_list, taxon_id, app_name=app_name, version=version, quiet=quiet)
    id_di = {}
    my_genes = []
    for line in range(len(results.index)):
        input_identifier, string_identifier = results["queryItem"][line], results["stringId"][line]
        my_genes.append(string_identifier)
        id_di.setdefault(string_identifier, input_identifier)
    if (not quiet):
        print("<STRING> Getting the STRING network interactions")
    output_format = "tsv"
    method = "network"
    request_url = "/".join([string_api_url(version), output_format, method])
    params = {
        "identifiers" : "%0d".join(my_genes), # your protein
        "species" : taxon_id, # species NCBI identifier 
        "limit" : limit, # limits the number of interaction partners retrieved per protein
        "required_score" : int(min_score*1000), # in 0 - 1000, 0 : get all edges
        "network_type": network_type,
        "caller_identity" : app_name # your app name
    }
    if (not quiet):
        print("<STRING> Getting the STRING network interactions")
    response = requests.post(request_url, data=params).text
    sleep(1)
    network = pd.read_csv(StringIO(response), sep="\t")
    network["preferredName_A"] = [id_di.get(x, x) for x in list(network["preferredName_A"])]
    network["preferredName_B"] = [id_di.get(x, x) for x in list(network["preferredName_B"])]
    network = network[["preferredName_A","preferredName_B","score"]]
    network["sign"] = [2]*network.shape[0]
    network["directed"] = [0]*network.shape[0]
    network = network[["preferredName_A","preferredName_B","sign","directed","score"]]
    network = network.drop_duplicates(keep="first")
    network.sort_values(by="score", ascending=False)
    return network

def get_interactions_from_STRING(gene_list, taxon_id, min_score=0, app_name=None, file_folder=None, version="11.0", strict=False, quiet=False):
    '''
    Retrieves (un)directed and (un)signed physical interactions from the STRING database

    ...

    Parameters
    ----------
    gene_list : Python character string list
        list of genes
    taxon_id : Python integer
        NCBI taxonomy ID
    min_score : Python integer
        [default=0] : in [0,1] STRING combined score
    app_name : Python character string
        [default=None] : identifier for STRING
    file_folder : Python character string
        [default=None]: where to save the file from STRING (if None, the file is not saved)
    version : Python character string
        [default="v11.0"] : STRING database version
    strict : Python bool
        [default=False] : if set to True, only keep interactions involving genes BOTH in @gene_list
    quiet : Python bool
        [default=False] : prints out verbose

    Returns
    ----------
    res_df : Pandas Dataframe
        rows/[interation number] x columns/["preferredName_A", "preferredName_B", "sign", "directed", "score"]
    '''
    assert version!="11.5"
    assert app_name
    assert taxon_id
    assert min_score<=1 and min_score>=0
    protein_action_fname = (file_folder if (file_folder) else "")+"protein_action_v"+version+".tsv"
    species = str(taxon_id)
    ftype ="protein.actions"
    fname_ = ftype+".v"+version+".txt"
    alpha, beta = version.split(".")
    STRING_url = "https://version-"+alpha+"-"+beta+".string-db.org/download/"+ftype+".v"+version+"/"+species+"."+fname_+".gz"
    if (not quiet):
        print("<STRING> Retrieving the file from STRING", end="... ")
    if (file_folder is not None and not os.path.exists(protein_action_fname)):
        sbcall("wget -qO- \""+STRING_url+"\" | gzip -d -c > "+protein_action_fname, shell=True)
    if (file_folder is not None):
        if (not quiet):
            print("Saved at %s" % protein_action_fname)
        df = pd.read_csv(protein_action_fname, sep="\t")
    else:
        if (not quiet):
            print("Downloaded online")
        df = pd.read_csv(StringIO(sbcheck_output("wget -qO- \""+STRING_url+"\" | gzip -d -c", shell=True).decode("utf-8")), sep="\t")
    df = df.loc[df["score"]>=int(min_score*1000)]
    df["score"] /= 1000
    res_df = get_protein_names_from_STRING(gene_list, taxon_id, app_name=app_name, quiet=quiet)
    res_df.index = res_df["stringId"]
    res_df = res_df["preferredName"].to_dict()
    df1 = deepcopy(df)
    df1["item_id_a"] = [res_df.get(a, "-") for a in df1["item_id_a"]]
    df1["item_id_b"] = [res_df.get(b, "-") for b in df1["item_id_b"]]
    df = df.loc[(df1["item_id_a"]!="-")|(df1["item_id_b"]!="-")]
    genes = list(set(list(df["item_id_a"])+list(df["item_id_b"])))
    res_df = get_protein_names_from_STRING([x.split(str(taxon_id)+".")[-1] for x in genes], taxon_id, app_name=app_name, quiet=quiet)
    res_df.index = res_df["stringId"]
    res_df = res_df["preferredName"].to_dict()
    df["item_id_a"] = [res_df.get(a, "-") for a in df["item_id_a"]]
    df["item_id_b"] = [res_df.get(b, "-") for b in df["item_id_b"]]
    df = df.loc[(df["item_id_a"]!="-")&(df["item_id_b"]!="-")]
    if (not quiet):
        print("... Protein identifier matching")
    ## if sign is known
    df["sign"] = [{"activation": 1, "inhibition": -1}.get(x, 2) for x in df["action"]]
    if (not quiet):
        print("... Signed")
    ## if direction is known
    df["directed"] = [int(x=="t") for x in df["is_directional"]]
    df_A = df.loc[df["a_is_acting"]=="f"]
    df_B = df.loc[df["a_is_acting"]=="t"]
    if (not quiet):
        print("... Directed")
    network = pd.DataFrame([], index=range(df.shape[0]))
    network["preferredName_A"] = list(df_B["item_id_a"])+list(df_A["item_id_b"])
    network["preferredName_B"] = list(df_B["item_id_b"])+list(df_A["item_id_a"])
    network["sign"] = list(df_B["sign"])+list(df_A["sign"])
    network["directed"] = list(df_B["directed"])+list(df_A["directed"])
    network["score"] = list(df_B["score"])+list(df_A["score"])
    if (not quiet):
        print("... Aggregate info")
    ## remove redundant, solve paradoxes
    if (strict):
        f = np.vectorize(lambda x : x in gene_list)
        network = network.loc[f(network["preferredName_A"])&f(network["preferredName_B"])]
    network.index = ["--".join(x) for x in zip(list(network["preferredName_A"]), list(network["preferredName_B"]))]
    network = network.sort_values(by="score", ascending=False) #sort by decreasing score
    solve_conflicts_directed = {x: int(network.loc[x]["directed"].max()) for x in network.index} #if reported once directed, considered directed
    if (not quiet):
        print("... Solve conflicts on direction")
    is_unsigned = lambda sign : ((sign==2).all() or ((sign==-1).any() and (sign==1).any()))
    #for a fixed directed edge, if several different signs are reported, if both signs are present, consider it unsigned
    solve_conflicts_sign = {x: int(network.loc[x]["sign"]) if ("numpy.int64" in str(type(network.loc[x]["sign"]))) else (2 if (is_unsigned(network.loc[x]["sign"])) else (-1)**int((network.loc[x]["sign"]==-1).all())) for x in network.index}
    if (not quiet):
        print("... Solve conflicts on sign for directed edges")
    network = network.loc[~network.index.duplicated(keep="first")] #keep the duplicate with highest score
    network["directed"] = [solve_conflicts_directed[x] for x in network.index]
    network["sign"] = [solve_conflicts_sign[x] for x in network.index]
    final_indices = list(set([idx if (network.loc[idx]["directed"]==1) else ("--".join(list(sorted(idx.split("--"))))) for idx in network.index]))
    network = network.loc[final_indices]
    if (not quiet):
        print("... Aggregate info again")
    directed,indices=[],[]
    for idx in network.index:
        if (("--".join(list(sorted(idx.split("--"))))) in indices):
            continue
        if (("--".join(list(sorted(idx.split("--"))))) in network.index and (network.loc[idx]["directed"]==1)):
            directed.append(0)
        else:
            directed.append(network.loc[idx]["directed"])
        indices.append(idx)
    network = network.loc[indices]
    network["directed"] = directed
    if (not quiet):
        print("... Remove multiple undirected edges")
    #for a fixed UNDIRECTED edge, if several different signs are reported, if both signs are present, consider it unsigned
    network.index = ["--".join(list(sorted([network.loc[x]["preferredName_A"],network.loc[x]["preferredName_B"]]))) for x in network.index] 
    solve_conflicts_sign_undirected = {x: int(network.loc[x]["sign"]) if ((network.loc[x]["directed"]!=0).any() or ("numpy.int64" in str(type(network.loc[x]["sign"])))) else (2 if (is_unsigned(network.loc[x]["sign"])) else (-1)**int((network.loc[x]["sign"]==-1).all())) for x in network.index}
    if (not quiet):
        print("... Solve conflicts on sign for undirected edges")
    for ix, x in enumerate(network.index):
        values = list(network.iloc[ix,:].values.flatten()) 
        values[list(network.columns).index("sign")] = int(solve_conflicts_sign_undirected[x])
        values[list(network.columns).index("directed")] = int(values[list(network.columns).index("directed")])
        values[list(network.columns).index("score")] = float(values[list(network.columns).index("score")])
        network.loc[x] = values
    if (not quiet):
        print("... Aggregate info once again")
    network.index = range(network.shape[0])
    network.sort_values(by="score", ascending=False)
    return network


#coding: utf-8

import NORDic
import os
import pandas as pd
from multiprocessing import cpu_count
import numpy as np
from subprocess import call
import unittest

## I. Building a small regulatory model of CCHS (**NORDic NI**)

class TestNI(unittest.TestCase):

    def create_instance(self):
        ## Registration to databases
        DisGeNET_credentials = "../tests/credentials_DISGENET.txt"
        STRING_credentials = "../tests/credentials_STRING.txt"
        LINCS_credentials = "../tests/credentials_LINCS.txt"
        ## Parameters
        seed_number=123456
        njobs=max(1,cpu_count()-2) ## all available threads but 2
        file_folder="ToyOndine/"
        taxon_id=9606 # human species
        disease_cids=["C1275808"] ## Concept ID of Ondine syndrome
        cell_lines=["NPC", "SHSY5Y"] # brain cell lines in LINCS L1000
        ## Information about the disease
        DISGENET_args = {"credentials": DisGeNET_credentials, "disease_cids": disease_cids}
        ## Selection of parameters relative to the prior knowledge network 
        STRING_args = {"credentials": STRING_credentials, "score": 0}
        EDGE_args = {"tau": 0, "filter": True, "connected": True}
        accept_nonRNA=True
        preserve_network_sign=True
        ## Selection of parameters relative to experimental constraints
        LINCS_args = {"path_to_lincs": "../lincs/", "credentials": LINCS_credentials, "cell_lines": cell_lines, 
              "thres_iscale": None}
        SIG_args = {"bin_thres": 0.5}
        force_experiments=False
        ## Selection of parameters relative to the inference of networks
        BONESIS_args = {"limit": 1, "exact": True, "max_maxclause": 3}
        ## Advanced
        DESIRABILITY = {"DS": 3, "CL": 3, "Centr": 3, "GT": 1}
        call("mkdir -p "+file_folder, shell=True)

        ## The undirected, unsigned network from STRING
        network_content = pd.DataFrame([], index=["preferredName_A", "preferredName_B", "sign", "directed", "score"])
        network_content[0] = ["PHOX2B", "BDNF", 2, 0, 0.342]
        network_content[1] = ["PHOX2B", "GDNF", 2, 0, 0.572]
        network_content[2] = ["PHOX2B", "RET", 2, 0, 0.605]
        network_content[3] = ["PHOX2B", "EDN3", 2, 0, 0.607]
        network_content[4] = ["PHOX2B", "ASCL1", 2, 0, 0.676]
        network_content[5] = ["ASCL1", "RET", 2, 0, 0.397]
        network_content[6] = ["ASCL1", "EDN3", 2, 0, 0.433]
        network_content[7] = ["ASCL1", "GDNF", 2, 0, 0.47]
        network_content[8] = ["ASCL1", "BDNF", 2, 0, 0.519]
        network_content[9] = ["EDN3", "BDNF", 2, 0, 0.15]
        network_content[10] = ["EDN3", "RET", 2, 0, 0.622]
        network_content[11] = ["EDN3", "GDNF", 2, 0, 0.634]
        network_content[12] = ["RET", "BDNF", 2, 0, 0.438]
        network_content[12] = ["RET", "GDNF", 2, 0, 0.999]
        network_content[12] = ["GDNF", "BDNF", 2, 0, 0.95]
        network_content.T.to_csv(file_folder+"network.tsv", sep="\t", index=None)

        ## An experiment retrieved from LINCS L1000 involving those genes
        index=["PHOX2B","EDN3","RET","GDNF","BDNF","ASCL1","cell_line","annotation","perturbed","perturbation","sigid"]
        experiments_content = pd.DataFrame([], index=index)
        experiments_content["BDNF_KD_SHSY5Y"] = [1,1,1,0,0,0,"Cell","2","BDNF","KD","Sig1"] ## mutated profile
        experiments_content["initial_SHSY5Y"] = [1,1,1,1,0,0,"Cell","1","None","None","Sig2"] ## control/initial profile
        experiments_content.to_csv(file_folder+"experiments.csv")
        return file_folder, taxon_id, DISGENET_args, STRING_args, EDGE_args, LINCS_args, SIG_args, BONESIS_args,DESIRABILITY, seed_number, njobs, force_experiments, accept_nonRNA, preserve_network_sign

    def test_network_identification(self):
        file_folder, taxon_id, DISGENET_args, STRING_args, EDGE_args, LINCS_args, SIG_args, BONESIS_args,DESIRABILITY, seed_number, njobs, force_experiments, accept_nonRNA, preserve_network_sign = self.create_instance()
        ## Network identification
        from NORDic.NORDic_NI.functions import network_identification
        solution = network_identification(file_folder, taxon_id, path_to_genes=None, 
            network_fname=file_folder+"network.tsv", experiments_fname=file_folder+"experiments.csv", 
            disgenet_args=DISGENET_args, string_args=STRING_args, edge_args=EDGE_args, lincs_args=LINCS_args, 
            sig_args=SIG_args, bonesis_args=BONESIS_args, weights=DESIRABILITY, seed=seed_number, njobs=njobs,
            force_experiments=force_experiments, accept_nonRNA=accept_nonRNA, preserve_network_sign=preserve_network_sign)
        self.assertTrue(os.path.exists(file_folder+'inferred_max_criterion_solution.png'))
        with open(file_folder+"solution.bnet", "r") as f:
            network = f.read().split("\n")
        self.assertEqual(len(network),6)

if __name__ == '__main__':
    unittest.main()


#coding: utf-8

import NORDic
import os
import pandas as pd
from multiprocessing import cpu_count
import numpy as np
import unittest

## III. Simulation of drug effect based on a Boolean network (**NORDic DS**)

class TestDS(unittest.TestCase):

    def create_instance(self):
        file_folder="ToyOndine/"
        seed_number=12345
        njobs=min(5,max(1,cpu_count()-2))
        SIMU_params = {'nb_sims': 1000, 'rates': "fully_asynchronous", 'thread_count': njobs, 'depth': "constant_unitary"}
        np.random.seed(seed_number)
        with open(file_folder+"solution.bnet", "r") as f:
            genes = [line.split(", ")[0] for line in f.read().split("\n") if (len(line)>0)]
        ## Create a set of random "patient" and "control" states (see tests_PMR.py)
        states = pd.read_csv(file_folder+"phenotypes.csv", index_col=0)
        ### Patient profiles
        patients = states[[c for c in states.columns if ("Case" in c)]].loc[[i for i in states.index if (i != "annotation")]]
        ## Generate a set of drugs
        targets = {"KO_"+g: {g: -1} for g in genes}
        targets.update({"OE_"+g: {g: 1} for g in genes})
        targets = pd.DataFrame(targets)
        all_genes = pd.DataFrame({"One": {g: 0 for g in genes}})
        targets = targets.join(all_genes, how="outer").fillna(0)[[c for c in targets.columns if (c!="One")]].astype(int)
        targets.to_csv(file_folder+"targets.csv")
        return states, targets, patients, SIMU_params, file_folder, seed_number, njobs

    def create_frontier(self):
        states, targets, patients, SIMU_params, file_folder, seed_number, njobs = self.create_instance()
        ## Drug Scoring
        ### Define the function which determines whether a state is closer to controls/patients
        dfdata = states.loc[[i for i in states.index if (i != "annotation")]]
        samples = states.loc["annotation"]
        from NORDic.NORDic_DS.functions import compute_frontier
        ## compute_frontier fits a SVM model to the data
        frontier = compute_frontier(dfdata, samples)
        score = lambda attrs : (frontier.predict(attrs.values.T)==1).astype(int)
        return states, targets, patients, SIMU_params, file_folder, seed_number, njobs, score

    def test_drug_simulation(self):
        states, targets, patients, SIMU_params, file_folder, seed_number, njobs, score = self.create_frontier()
        from NORDic.NORDic_DS.functions import simulate
        scores = simulate(file_folder+"solution.bnet", targets, patients, score, simu_params=SIMU_params, nbseed=seed_number)
        self.assertEqual(scores.shape[0],patients.shape[1])
        self.assertEqual(scores.shape[1],targets.shape[1])

if __name__ == '__main__':
    unittest.main()


#coding: utf-8

import NORDic
import os
import pandas as pd
import unittest
from subprocess import Popen

from functools import reduce

class TestPKN(unittest.TestCase):

    def create_network(self):
        ## Retrieve networks from the STRING database
        from NORDic.UTILS.STRING_utils import get_network_from_STRING
        file_folder="ToyOndine/"
        core_genes = ["PHOX2B", "RET", "BDNF", "ASCL1", "EDN3", "GDNF"]
        taxon_id = 9606
        if (not os.path.exists(file_folder+"network.csv")):
            network = get_network_from_STRING(core_genes, taxon_id, min_score=0., network_type="functional",
                add_nodes=0, app_name="NORDic PKN", version="11.5", quiet=0)
            network.to_csv(file_folder+"network.csv")
        network = pd.read_csv(file_folder+"network.csv", index_col=0)
        return network

    def create_PPI(self):
        ## Retrieve networks from the STRING database
        from NORDic.UTILS.STRING_utils import get_interactions_from_STRING
        file_folder="ToyOndine/"
        core_genes = ["PHOX2B", "RET", "BDNF", "ASCL1", "EDN3", "GDNF"]
        taxon_id = 9606
        if (not os.path.exists(file_folder+"PPI.csv")):
            PPI = get_interactions_from_STRING(core_genes, taxon_id, min_score=0., strict=False,
                    version="11.0", app_name="NORDic PKN PPI", file_folder=file_folder)
            PPI.to_csv(file_folder+"PPI.csv")
        PPI = pd.read_csv(file_folder+"PPI.csv", index_col=0)
        return PPI

    def test_merge_network_PPI(self):
        ## Retrieve networks from the STRING database
        from NORDic.UTILS.utils_network import merge_network_PPI
        file_folder="ToyOndine/"
        core_genes = ["PHOX2B", "RET", "BDNF", "ASCL1", "EDN3", "GDNF"]
        taxon_id = 9606
        network = self.create_network()
        PPI = self.create_PPI()
        if (not os.path.exists(file_folder+"final_network.csv")):
            final_network = merge_network_PPI(network, PPI)
            final_network.to_csv(file_folder+"final_network.csv")
        final_network = pd.read_csv(file_folder+"final_network.csv", index_col=0)
        return final_network, core_genes

    def test_determine_threshold(self):
        from NORDic.UTILS.utils_network import determine_edge_threshold
        from NORDic.UTILS.utils_network import remove_isolated
        from NORDic.UTILS.utils_grn import get_weakly_connected
        from functools import reduce
        file_folder="ToyOndine/"
        final_network, core_genes = self.test_merge_network_PPI()
        threshold = determine_edge_threshold(final_network, core_genes)
        self.assertTrue(0 < threshold)
        self.assertTrue(threshold < 1)
        final_network = remove_isolated(final_network.loc[final_network["score"]>=threshold])
        glist = list(set(reduce(lambda x,y: x+y, [list(final_network[c]) for c in ["preferredName_A", "preferredName_B"]])))
        components = get_weakly_connected(final_network, glist, score_col="score")
        self.assertEqual(len(components),1)
        self.assertEqual(len(glist),len(components[0]))
        self.assertTrue(all([g in components[0] for g in core_genes]))
        final_network.index = range(final_network.shape[0])
        NETWORK_fname = file_folder+"network.tsv"
        final_network.to_csv(NETWORK_fname, sep="\t", index=None)
        return final_network

    def test_plot_influence_graph(self):
        from NORDic.UTILS.utils_plot import plot_influence_graph
        file_folder="ToyOndine/"
        final_network = self.test_determine_threshold()
        plot_influence_graph(final_network, "preferredName_A", "preferredName_B", "sign", direction_col="directed", fname=file_folder+"graph_final", optional=True)

    def test_full_pipeline(self):
        from NORDic.UTILS.STRING_utils import get_network_from_STRING
        from NORDic.UTILS.utils_network import aggregate_networks
        from NORDic.UTILS.utils_grn import get_weakly_connected
        file_folder="ToyOndine_full/"
        Popen(('cp -r ToyOndine/ '+file_folder).split(" "))
        core_genes = ["PHOX2B", "RET", "BDNF", "ASCL1", "EDN3", "GDNF"]
        taxon_id = 9606
        network = get_network_from_STRING(core_genes, taxon_id, min_score=0., network_type="functional",
                add_nodes=0, app_name="NORDic PKN", version="11.5", quiet=0)
        network.to_csv(file_folder+"network.csv")
        final_network2 = aggregate_networks(file_folder, core_genes, taxon_id, 0, "functional", "NORDic whole PKN", quiet=1)
        glist2 = list(set(reduce(lambda x,y: x+y, [list(final_network2[c]) for c in ["preferredName_A", "preferredName_B"]])))
        components2 = get_weakly_connected(final_network2, glist2, score_col="score")
        self.assertEqual(len(components2),1)
        self.assertEqual(len(glist2),len(components2[0]))
        self.assertTrue(all([g in components2[0] for g in core_genes]))

if __name__ == '__main__':
    unittest.main()


#coding: utf-8

import NORDic
import pandas as pd
import numpy as np
import os
from multiprocessing import cpu_count
from time import time
import unittest

class TestSIM(unittest.TestCase):

    def create_instance(self):
        taxon_id = 9606
        njobs=max(1,cpu_count()-2)
        seednb=12345
        file_folder="ToyOndine/"
        network_fname=file_folder+"solution.bnet"
        import mpbn
        net = mpbn.MPBooleanNetwork(mpbn.load(network_fname))
        x0 = net.zero() # consider initial state where all gene expression levels are set to 0...
        for g in ['BDNF', 'EDN3']: #... except for 'BDNF' and 'EDN3'
            x0[g] = 1
        return x0, network_fname, seednb, njobs

    def test_MABOSS_SIM(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        from NORDic.UTILS.utils_sim import MABOSS_SIM
        initial = pd.DataFrame(pd.Series(x0), columns=["initial"])
        params = {'sample_count': 10000, 'max_time': 100}
        nee = MABOSS_SIM(seednb,njobs)
        nee.update_network(network_fname, initial, verbose=False)
        probs = nee.generate_trajectories(params=params, outputs=[]) ## generate trajectories
        self.assertTrue(all([c=="<nil>" for c in probs.columns]))
        self.assertTrue(np.isclose(probs.loc["prob"]["<nil>"],1))

    def test_MPBN_SIM(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        from NORDic.UTILS.utils_sim import MPBN_SIM
        initial = pd.DataFrame(pd.Series(x0), columns=["initial"])
        params = {'sample_count': 10000, 'max_time': 100}
        nee = MPBN_SIM(seednb,njobs)
        nee.update_network(network_fname, initial, verbose=False) ## no mutations, initial state x0
        probs = nee.generate_trajectories(params=params, outputs=[]) ## generate trajectories
        self.assertTrue(all([c=="<nil>" for c in probs.columns]))
        self.assertTrue(np.isclose(probs.loc["prob"]["<nil>"],1))

    def test_BONESIS_SIM_FinalEmpty(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        from NORDic.UTILS.utils_sim import BONESIS_SIM
        initial = pd.DataFrame(pd.Series(x0), columns=["initial"])
        final = pd.DataFrame([], index=[g for g in x0], columns=["final"]).fillna(0) 
        nee = BONESIS_SIM(seednb,njobs)
        nee.update_network(network_fname, initial, final, verbose=False) ## no mutations, initial state x0
        attrs = nee.enumerate_attractors(verbose=False) 
        self.assertTrue("StateNotFound" not in attrs.columns[0])

    def test_BONESIS_SIM_FinalState(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        from NORDic.UTILS.utils_sim import BONESIS_SIM
        initial = pd.DataFrame(pd.Series(x0), columns=["initial"])
        final = pd.DataFrame([[1,0,1,1,1,1]], columns=[g for g in x0], index=["final"]).T 
        nee = BONESIS_SIM(seednb,njobs)
        nee.update_network(network_fname, initial, final, verbose=False) ## no mutations, initial state x0
        attrs = nee.enumerate_attractors(verbose=False) 
        self.assertTrue("StateNotFound" in attrs.columns[0])

    def test_MABOSS_SIM_Attractors(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        from NORDic.UTILS.utils_sim import MABOSS_SIM
        initial = pd.DataFrame(pd.Series(x0), columns=["initial"])
        nee = MABOSS_SIM(seednb,njobs)
        nee.update_network(network_fname, initial, verbose=False)
        nee.enumerate_attractors() ## enumerate all attractors
        self.assertEqual(nee.attrs.shape[1], 1)
        self.assertTrue(all(["".join(list(nee.attrs.astype(str)[c])) in ["1"*len(nee.gene_list), "0"*len(nee.gene_list)] for c in nee.attrs]))

    def test_MPBN_SIM_Attractors(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        from NORDic.UTILS.utils_sim import MPBN_SIM
        initial = pd.DataFrame(pd.Series(x0), columns=["initial"])
        nee = MPBN_SIM(seednb,njobs)
        nee.update_network(network_fname, initial, verbose=False)
        nee.enumerate_attractors() ## enumerate all attractors
        self.assertEqual(nee.attrs.shape[1],1)
        self.assertTrue(all(["".join(list(nee.attrs.astype(str)[c])) in ["1"*len(nee.gene_list), "0"*len(nee.gene_list)] for c in nee.attrs]))

    def test_MABOSS_Trajectories_Profiles(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        C = pd.DataFrame(pd.Series(x0), columns=["initial"])
        T = pd.DataFrame([[1,0,1,1,1,1]], columns=[g for g in x0], index=["NotReachable"]).T 
        T = T.join(pd.DataFrame([], index=[g for g in x0], columns=["Reachable"]).fillna(0), how="outer")
        from NORDic.UTILS.utils_sim import test
        from NORDic.UTILS.utils_sim import MABOSS_SIM
        _, net_MABOSS_noMut = test(MABOSS_SIM,seednb,njobs,network_fname,C,T,"profiles",verbose=0)
        res = pd.DataFrame({"score": net_MABOSS_noMut.max_values}).T
        self.assertTrue(np.isclose(res.loc["score"]["initial->Reachable"],1))
        self.assertTrue(res.loc["score"]["initial->NotReachable"]<1)

    def test_MABOSS_Trajectories_Attractors(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        C = pd.DataFrame(pd.Series(x0), columns=["initial"])
        T = pd.DataFrame([[1,0,1,1,1,1]], columns=[g for g in x0], index=["NotReachable"]).T 
        T = T.join(pd.DataFrame([], index=[g for g in x0], columns=["Reachable"]).fillna(0), how="outer")
        from NORDic.UTILS.utils_sim import test
        from NORDic.UTILS.utils_sim import MABOSS_SIM
        _, net_MABOSS_noMut = test(MABOSS_SIM,seednb,njobs,network_fname,C,T,"attractors",verbose=0)
        res = pd.DataFrame({"score": net_MABOSS_noMut.max_values}).T
        self.assertTrue(np.isclose(res.loc["score"]["initial->Reachable"],1))
        self.assertTrue(np.isclose(res.loc["score"]["initial->NotReachable"],1))

    def test_MPBN_Trajectories_Profiles(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        C = pd.DataFrame(pd.Series(x0), columns=["initial"])
        T = pd.DataFrame([[1,0,1,1,1,1]], columns=[g for g in x0], index=["NotReachable"]).T 
        T = T.join(pd.DataFrame([], index=[g for g in x0], columns=["Reachable"]).fillna(0), how="outer")
        from NORDic.UTILS.utils_sim import test
        from NORDic.UTILS.utils_sim import MPBN_SIM
        _, net_MPBN_noMut = test(MPBN_SIM,seednb,njobs,network_fname,C,T,"profiles",verbose=0)
        res = pd.DataFrame({"score": net_MPBN_noMut.max_values}).T
        self.assertTrue(np.isclose(res.loc["score"]["initial->Reachable"],1))
        self.assertTrue(res.loc["score"]["initial->NotReachable"]<1)

    def test_MPBN_Trajectories_Attractors(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        C = pd.DataFrame(pd.Series(x0), columns=["initial"])
        T = pd.DataFrame([[1,0,1,1,1,1]], columns=[g for g in x0], index=["NotReachable"]).T 
        T = T.join(pd.DataFrame([], index=[g for g in x0], columns=["Reachable"]).fillna(0), how="outer")
        from NORDic.UTILS.utils_sim import test
        from NORDic.UTILS.utils_sim import MPBN_SIM
        _, net_MPBN_noMut = test(MPBN_SIM,seednb,njobs,network_fname,C,T,"attractors",verbose=0)
        res = pd.DataFrame({"score": net_MPBN_noMut.max_values}).T
        self.assertTrue(np.isclose(res.loc["score"]["initial->Reachable"],1))
        self.assertTrue(np.isclose(res.loc["score"]["initial->NotReachable"],1))

    def test_BONESIS_Trajectories_Profiles(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        C = pd.DataFrame(pd.Series(x0), columns=["initial"])
        T = pd.DataFrame([[1,0,1,1,1,1]], columns=[g for g in x0], index=["NotReachable"]).T 
        T = T.join(pd.DataFrame([], index=[g for g in x0], columns=["Reachable"]).fillna(0), how="outer")
        from NORDic.UTILS.utils_sim import test
        from NORDic.UTILS.utils_sim import BONESIS_SIM
        _, net_BONESIS_noMut = test(BONESIS_SIM,seednb,njobs,network_fname,C,T,"profiles",verbose=0)
        res = pd.DataFrame({"score": net_BONESIS_noMut.max_values}).T
        self.assertTrue(np.isclose(res.loc["score"]["initial->Reachable"],1))
        self.assertTrue(np.isclose(res.loc["score"]["initial->NotReachable"],0))

    def test_BONESIS_Trajectories_Attractors(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        C = pd.DataFrame(pd.Series(x0), columns=["initial"])
        T = pd.DataFrame([[1,0,1,1,1,1]], columns=[g for g in x0], index=["NotReachable"]).T 
        T = T.join(pd.DataFrame([], index=[g for g in x0], columns=["Reachable"]).fillna(0), how="outer")
        from NORDic.UTILS.utils_sim import test
        from NORDic.UTILS.utils_sim import BONESIS_SIM
        _, net_BONESIS_noMut = test(BONESIS_SIM,seednb,njobs,network_fname,C,T,"attractors",verbose=0)
        res = pd.DataFrame({"score": net_BONESIS_noMut.max_values}).T
        self.assertTrue(np.isclose(res.loc["score"]["initial->Reachable"],1))
        self.assertTrue(np.isclose(res.loc["score"]["initial->NotReachable"],0))

    def test_MPBN_Mutant_KO(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        initial = pd.DataFrame(pd.Series(x0), columns=["initial"])
        from NORDic.UTILS.utils_sim import test
        from NORDic.UTILS.utils_sim import MPBN_SIM
        nee = MPBN_SIM(seednb,njobs)
        nee.update_network(network_fname, initial, mutation_permanent={'PHOX2B': 0}, verbose=False)
        self.assertEqual(len(nee.all_mutants),1)
        self.assertTrue('PHOX2B' in nee.all_mutants)

    def test_MPBN_Mutant_OE(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        initial = pd.DataFrame(pd.Series(x0), columns=["initial"])
        from NORDic.UTILS.utils_sim import test
        from NORDic.UTILS.utils_sim import MPBN_SIM
        nee = MPBN_SIM(seednb,njobs)
        nee.update_network(network_fname, initial, mutation_permanent={'PHOX2B': 1}, verbose=False)
        self.assertEqual(len(nee.all_mutants),1)
        self.assertTrue('PHOX2B' in nee.all_mutants)

    def test_MPBN_Mutant_KO_transient(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        initial = pd.DataFrame(pd.Series(x0), columns=["initial"])
        from NORDic.UTILS.utils_sim import test
        from NORDic.UTILS.utils_sim import MPBN_SIM
        nee = MPBN_SIM(seednb,njobs)
        nee.update_network(network_fname, initial, mutation_transient={'PHOX2B': 0}, verbose=False)
        self.assertEqual(len(nee.all_mutants),1)
        self.assertTrue('PHOX2B' in nee.all_mutants)

    def test_MPBN_Mutant_OE_transient(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        initial = pd.DataFrame(pd.Series(x0), columns=["initial"])
        from NORDic.UTILS.utils_sim import test
        from NORDic.UTILS.utils_sim import MPBN_SIM
        nee = MPBN_SIM(seednb,njobs)
        nee.update_network(network_fname, initial, mutation_transient={'PHOX2B': 1}, verbose=False)
        self.assertEqual(len(nee.all_mutants),1)
        self.assertTrue('PHOX2B' in nee.all_mutants)

    def test_MPBN_Trajectory_Profiles_Mutant_KO(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        C = pd.DataFrame(pd.Series(x0), columns=["initial"])
        T = pd.DataFrame([[1,0,1,1,1,1]], columns=[g for g in x0], index=["NotReachable"]).T 
        T = T.join(pd.DataFrame([], index=[g for g in x0], columns=["Reachable"]).fillna(0), how="outer")
        from NORDic.UTILS.utils_sim import test
        from NORDic.UTILS.utils_sim import MPBN_SIM
        _, net_MPBN_noMut = test(MPBN_SIM,seednb,njobs,network_fname,C,T,"profiles", mutation_permanent={'PHOX2B': 0},verbose=0)
        res = pd.DataFrame({"score": net_MPBN_noMut.max_values}).T
        self.assertTrue(np.isclose(res.loc["score"]["initial->Reachable"],1))
        self.assertTrue(res.loc["score"]["initial->NotReachable"]<1)

    def test_MPBN_Trajectory_Profiles_Mutant_KO_transient(self):
        x0, network_fname, seednb, njobs = self.create_instance()
        C = pd.DataFrame(pd.Series(x0), columns=["initial"])
        T = pd.DataFrame([[1,0,1,1,1,1]], columns=[g for g in x0], index=["NotReachable"]).T 
        T = T.join(pd.DataFrame([], index=[g for g in x0], columns=["Reachable"]).fillna(0), how="outer")
        from NORDic.UTILS.utils_sim import test
        from NORDic.UTILS.utils_sim import MPBN_SIM
        _, net_MPBN_noMut = test(MPBN_SIM,seednb,njobs,network_fname,C,T,"profiles", mutation_transient={'EDN3': 0},verbose=0)
        res = pd.DataFrame({"score": net_MPBN_noMut.max_values}).T
        self.assertTrue(np.isclose(res.loc["score"]["initial->Reachable"],1))
        self.assertTrue(res.loc["score"]["initial->NotReachable"]<1)

if __name__ == '__main__':
    unittest.main()